// gcc -std=c11 -Wall -Wextra bash_shock_fix.c -o bash_shock_fix

// For usage and license, see https://ad5ey.net/bash_shock_fix.html

#include <sys/types.h>
#include <regex.h>
#include <stdio.h>	// perror
#include <stdlib.h>	// calloc, free
#include <syslog.h>	// syslog
#include <unistd.h>	// execvpe
#include <string.h>	// memset, strlen

//LOG_EMERG      system is unusable
//LOG_ALERT      action must be taken immediately
//LOG_CRIT       critical conditions
//LOG_ERR        error conditions
//LOG_WARNING    warning conditions
//LOG_NOTICE     normal, but significant, condition
//LOG_INFO       informational message
//LOG_DEBUG      debug-level message

void print_regex_error(const char *prefix, int reg_errcode, const regex_t *preg) {
	size_t errbuf_size = regerror(reg_errcode, preg, 0, 0);
	char *errbuf=calloc(errbuf_size, 1);
	if(!errbuf) {
		syslog(LOG_CRIT, "error: calloc failed to allocate %zu bytes\n", errbuf_size);
		return;
	}
	size_t errbuf_size_check = regerror(reg_errcode, preg, errbuf, errbuf_size);
	errbuf[errbuf_size-1]=0;	// guarantee null termination
	if (errbuf_size_check!=errbuf_size) {
		syslog(LOG_ERR, "warning: errbuf_size_check!=errbuf_size -- The following error message may be truncated.\n");
	}
	syslog(LOG_ERR, "%s: %.*s\n", prefix, (int)errbuf_size, errbuf);
	free(errbuf);
}

int main(int argc, char **argv, char **envp) {
	// Look for "= ( ) {" with any amount (or none) of whitespace between each of the 4 symbols.
	regex_t reg;
	int reg_errcode = regcomp(&reg, "[^=]=\\([[:space:]]*([[:space:]]*)[[:space:]]*{\\)", /*int cflags*/ 0 );
	if (reg_errcode) {
		print_regex_error("regcomp", reg_errcode, &reg);
		return 1;
	}
	for (unsigned ndx=0; envp[ndx]; ndx++) {
		regmatch_t match[2];
		memset(&match[0], 0, sizeof(regmatch_t)*2);
		reg_errcode = regexec(&reg, envp[ndx], 2, match, 0);

		switch (reg_errcode) {
			case 0:
			{
				size_t N = strlen(envp[ndx]);
				// Found a match.  We should log this, and prune this entry for now.
				// TODO: Should I only prune for certain user ID's?
				if (match[1].rm_so!=-1) {
					syslog(LOG_NOTICE, "Detected attack envp[%u] %i %i %i %i : '%s'\n", ndx, match[0].rm_so, match[0].rm_eo, match[1].rm_so, match[1].rm_eo, envp[ndx]);
					if ((size_t)match[1].rm_so>=N)
						match[1].rm_so = N-1;
					// Truncate the variable:
					envp[ndx][match[1].rm_so] = 0;
					syslog(LOG_NOTICE, "Deflected attack via truncating envp[%u]: '%s'\n", ndx, envp[ndx]);
				}
				else {
					syslog(LOG_ALERT, "FAILED TO DEFLECT ATTACK envp[%u] %i %i %i %i : '%s'\n", ndx, match[0].rm_so, match[0].rm_eo, match[1].rm_so, match[1].rm_eo, envp[ndx]);
				}
			}
			case REG_NOMATCH:
				break;	// Good.
			default:
				syslog(LOG_ERR, "Failed to clean envp[%u]: '%s'\n", ndx, envp[ndx]);
				print_regex_error("regexec", reg_errcode, &reg);
				regfree(&reg);
				return 1;
		}

	}       
	regfree(&reg);
	int ret = execvpe("/bin/bash.real", argv, envp);
	// We shouldn't be here. (execvpe should've replaced our process with bash)
	syslog(LOG_ERR, "execvpe: %m");
	return ret;
}
